#!/usr/bin/python3
#
#  libzvbi network identification example.
#
#  Copyright (C) 2006 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Example for the use of class Zvbi.ServiceDec, type
#   Zvbi.VBI_EVENT_NETWORK. This script shows how to identify a network
#   from data transmitted in XDS packets, Teletext packet 8/30 format 1
#   and 2, and VPS packets. The script captures from a device until the
#   currently tuned channel is identified by means of VPS, PDC et.al.
#
#   (This is a direct translation of examples/network.c in libzvbi.)

import sys
import argparse
import Zvbi

cap = None
vtdec = None
quit = False
opt = None

def ev_handler(ev_type, ev, user_data=None):
    # VBI_EVENT_NETWORK_ID is always sent when the decoder
    # receives a CNI. VBI_EVENT_NETWORK only if it can
    # determine a network name.

    if ev_type == Zvbi.VBI_EVENT_NETWORK:
        event_name = "VBI_EVENT_NETWORK"
    elif ev_type == Zvbi.VBI_EVENT_NETWORK_ID:
        event_name = "VBI_EVENT_NETWORK_ID"
    else:
        raise Exception("Unexpected event type: " + str(ev_type))

    if ev.name != "":
        # The network name is an ISO-8859-1 string (the API is quite old...) so
        # we convert it to locale encoding, nowadays usually UTF-8.
        # (Note this is done automatically within the Zvbi module)
        network_name = ev.name
    else:
        network_name = "unknown"

    # ASCII.
    if ev.call != "":
        call_sign = ev.call
    else:
        call_sign = "unknown"

    print(("%s: receiving: \"%s\" call sign: \"%s\" " +
           "CNI VPS: 0x%x 8/30-1: 0x%x 8/30-2: 0x%x") %
           (event_name,
            network_name,
            call_sign,
            ev.cni_vps,
            ev.cni_8301,
            ev.cni_8302))

    global quit
    if (((ev.cni_vps != 0) and opt.vps) or
        ((ev.cni_8301 != 0) and opt.p8301) or
        ((ev.cni_8302 != 0) and opt.p8302)):
        quit = True


def mainloop(services):
    # Should receive a CNI within two seconds, call sign within ten seconds(?).
    if (services & Zvbi.VBI_SLICED_CAPTION_525):
        n_frames = 11 * 30
    else:
        n_frames = 3 * 25

    for foo in range(0, n_frames):
        try:
            sliced_buf = cap.pull_sliced(10)
            vtdec.decode(sliced_buf)
        except Zvbi.CaptureError as e:
            print("Capture error:", e)
        except Zvbi.CaptureTimeout:
            pass

        if quit:
            return

    print("No network ID received or network unknown.")


def main_func():
    global cap
    global vtdec
    global opt

    if not opt.vps and not opt.p8301 and not opt.p8302:
        opt.vps = opt.p8301 = opt.p8302 = True

    services = (Zvbi.VBI_SLICED_TELETEXT_B |
                Zvbi.VBI_SLICED_VPS |
                Zvbi.VBI_SLICED_CAPTION_525)

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_buf_count = 5
        opt_strict = 0
        cap = Zvbi.Capture.Analog(opt.device, services=services,
                                  buffers=opt_buf_count, strict=opt_strict, trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    vtdec = Zvbi.ServiceDec()
    vtdec.event_handler_register( (Zvbi.VBI_EVENT_NETWORK |
                                   Zvbi.VBI_EVENT_NETWORK_ID), ev_handler )

    mainloop(services)


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='Capture and print network identification')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="VBI channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--vps", action='store_true', default=False, help="Use VPS only")
    parser.add_argument("--p8301", action='store_true', default=False, help="Use packet 8/30/1 only")
    parser.add_argument("--p8302", action='store_true', default=False, help="Use packet 8/30/2 only")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are multually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    pass



#!/usr/bin/python3
#
#  libzvbi test of teletext search
#
#  Copyright (C) 2007,2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# Description:
#
#   Example for the use of class Zvbi.Search. The script captures and
#   caches teletext pages until the RETURN key is pressed, then prompts
#   for a search string. A search across teletext pages is started, and
#   the content of matching pages is printed on the terminal.

import sys
import select
import argparse
import Zvbi

cr = '\r' if sys.stdout.isatty() else "\n"
last_pgno = -1
last_subno = -1

def pg_handler(pgtype, ev, user_data=None):
    print("%sPage %03x.%02x "
            % (cr, ev.pgno, ev.subno & 0xFF),
          file=sys.stderr, end='')


def progress(pg, pat):
    (page, sub) = pg.get_page_no()
    print("%sSearching %03x.%02X for \"%s\" "
            % (cr, page, sub, pat),
          file=sys.stderr, end='')
    return True  # let search continue


def search(vtdec):
    global last_pgno
    global last_subno

    pat = input("\nEnter search pattern: ")
    if pat:
        #rand = [int(rand(1000))]
        #print "Search rand user data rand->[0]\n"
        srch = Zvbi.Search(vtdec, pat, casefold=False, regexp=False,
                           progress=progress, userdata=pat)
  
        for pg in srch:
            (pgno, subno) = pg.get_page_no()
            if (last_pgno < 0) or (pgno != last_pgno) or (subno != last_subno):
                print("Found match: %03X.%04X\n" % (pgno, subno))
                print(pg.print_page(fmt='ascii').decode('utf-8'))
                last_pgno = pgno
                last_subno = subno

        print("")


def main_func():
    global cap
    global vtdec

    if opt.v4l2 or (opt.pid == 0 and not "dvb" in opt.device):
        opt_buf_count = 5
        opt_strict = 0
        cap = Zvbi.Capture.Analog(opt.device, services=services,
                                  buffers=opt_buf_count, strict=opt_strict, trace=opt.verbose)
    else:
        cap = Zvbi.Capture.Dvb(opt.device, dvb_pid=opt.pid, trace=opt.verbose)

    vtdec = Zvbi.ServiceDec()
    vtdec.event_handler_register(Zvbi.VBI_EVENT_TTX_PAGE, pg_handler)

    print("Press RETURN to stop capture and enter a search pattern", file=sys.stderr)

    # TODO use separate thread for capturing (as example & to avoid blocking during search)
    while True:
        # check if STDIN is readable (i.e. RETURN key was pressed)
        ret, foo1, foo2 = select.select([sys.stdin], [], [], 0)
        if ret:
            # discard one line of input (i.e. empty line caused by RETURN key)
            input()
            # ask for search string & perform search
            search(vtdec)

        try:
            sliced_buf = cap.pull_sliced(1000)
            vtdec.decode(sliced_buf)
        except Zvbi.CaptureError as e:
            print("Capture error:", e)
        except Zvbi.CaptureTimeout:
            print("Capture timeout:")


def ParseCmdOptions():
    global opt
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--v4l2", action='store_true', default=False, help="Using analog driver interface")
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.v4l2 and (opt.pid != 0):
        print("Options --v4l2 and --pid are multually exclusive", file=sys.stderr)
        sys.exit(1)
    if not opt.v4l2 and (opt.pid == 0) and ("dvb" in opt.device):
        print("WARNING: DVB devices require --pid parameter", file=sys.stderr)


try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    print("")  # needed due to progress output without trailing NL



#!/usr/bin/python3
#
#  Copyright (C) 2020 T. Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# Description:
#
#   Very simple test of the DVB PES de-multiplexer:
#   - reading stream from a DVB demux device
#   - calling de-multiplexer for each chunk of data
#   - forward sliced data
#
#   NOTE: Instead of reading from the device directly (and performing
#   device ioctl with hard-coded constants, which is prone to breakage)
#   you should use the Zvbi.Capture class which does that work for you.
#   Using the DvbDemux class directly only is useful when receiving a
#   stream from other sources (e.g. via socket from proxy on a remote
#   host).

import sys
import fcntl
import struct
import argparse
import Zvbi

#
# Callback invoked by DvbDemux.feed()
#
def sliced_handler(sliced_buf, pts, user_data=None):
    if opt.dump_sliced:
        ttx_lines = vps_lines = wss_lines = cc_lines = 0
        for (data, slc_id, line) in sliced_buf:
            if (slc_id & Zvbi.VBI_SLICED_TELETEXT_B):
                ttx_lines += 1
            elif (slc_id & Zvbi.VBI_SLICED_VPS):
                vps_lines += 1
            elif (slc_id & Zvbi.VBI_SLICED_WSS_625):
                wss_lines += 1
            elif (slc_id & (Zvbi.VBI_SLICED_CAPTION_625 | Zvbi.VBI_SLICED_CAPTION_525)):
                cc_lines += 1

        print("Received frame %.2f with %d lines: %d TTX, %d VPS, %d WSS, %d CC"
                % (pts, len(sliced_buf),
                   ttx_lines, vps_lines, wss_lines, cc_lines),
                file=sys.stderr)

    if opt.sliced:
        binary_sliced(sliced_buf)

    # return True so that feed function does not abort
    return True


#
# Dump packets in binary format equivalent examples/capture.py,
# thus suitable for reading by examples/decode.py
#
ServiceWidth = {
        Zvbi.VBI_SLICED_TELETEXT_B:  (42, 0),
        Zvbi.VBI_SLICED_CAPTION_625: (2, 1),
        Zvbi.VBI_SLICED_VPS:         (13, 2),
        Zvbi.VBI_SLICED_WSS_625:     (2, 3),
        Zvbi.VBI_SLICED_WSS_CPR1204: (3, 4),
        Zvbi.VBI_SLICED_CAPTION_525: (2, 7),
}

def binary_sliced(sliced_buf):
    outfile.write(bytes("%f\n" % sliced_buf.timestamp, 'ascii'))

    outfile.write(bytes([len(sliced_buf)]))

    for data, slc_id, line in sliced_buf:
        if ServiceWidth.get(slc_id) and (ServiceWidth.get(slc_id)[0] > 0):
            outfile.write(bytes([ServiceWidth.get(slc_id)[1],
                                 line & 0xFF,
                                 line >> 8]))
            data_len = ServiceWidth.get(slc_id)[0]
            outfile.write(data[0 : data_len])

    outfile.flush()


def main_func():
    if opt.use_callback:
        demux = Zvbi.DvbDemux(callback=sliced_handler)
    else:
        demux = Zvbi.DvbDemux()

    # open the DVB demux device
    try:
        dev = open(opt.device, "rb")
    except OSError as e:
        print("Failed to open device:", e, file=sys.stderr)
        sys.exit(1)

    # Configuring the PID in the DVB driver (in a very dirty way)
    # The following ioctl() call is equivalent this C code (Linux kernel 4.15.0):
    #   struct dmx_pes_filter_params filter
    #   filter.pid      = pid;
    #   filter.input    = DMX_IN_FRONTEND;
    #   filter.output   = DMX_OUT_TAP;
    #   filter.pes_type = DMX_PES_OTHER;
    #   filter.flags    = DMX_IMMEDIATE_START;
    #   ioctl (fd, DMX_SET_PES_FILTER, &filter))
    try:
        darg = struct.pack("@H2xIIII", opt.pid, 0, 1, 20, 4)
        fcntl.ioctl(dev.fileno(), 0x40146F2C, darg)
    except OSError as e:
        print("Failed to configure device via ioctl DMX_SET_PES_FILTER:", e, file=sys.stderr)
        sys.exit(1)

    # prepare for writing binary data to STDOUT
    if opt.sliced:
        global outfile
        outfile = open(sys.stdout.fileno(), "wb")

    # start reading data from the DVB device and demultiplex it
    # note the read function blocks until data is available
    while True:
        buf = dev.read(1024 * 8)
        if len(buf) == 0: break

        demux.feed(buf)

        if not opt.use_callback:
            for sliced_buf in demux:
                sliced_handler(sliced_buf, sliced_buf.timestamp)


def ParseCmdOptions():
    global opt
    parser = argparse.ArgumentParser(description='DVB de-multiplexer example')
    parser.add_argument("--device", type=str, default="/dev/dvb/adapter0/demux0", help="Path to video capture device")  # dev_name,
    parser.add_argument("--pid", type=int, default=0, help="Teletext channel PID for DVB")
    parser.add_argument("--use-callback", action='store_true', default=False, help="Use feed/callback API of DvbDemux class")
    parser.add_argument("--dump-sliced", action='store_true', default=False, help="Capture and all VBI services")
    parser.add_argument("--sliced", action='store_true', default=False, help="Write binary output, for piping into decode.py")   # bin_sliced,
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace output in the library")
    opt = parser.parse_args()

    if opt.pid <= 0:
        print("Warning: without valid --pid parameter no reception is possible", file=sys.stderr)


# main
try:
    ParseCmdOptions()
    main_func()
except (KeyboardInterrupt, BrokenPipeError):
    pass



#!/usr/bin/python3
#
#  Rudimentary render code for Closed Caption (CC) test.
#
#  Copyright (C) 2000, 2001 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#
# Description:
#
#   Example for the use of class Zvbi.ServiceDec, type Zvbi.VBI_EVENT_CAPTION.
#   When called without an input stream, the application opens a GUI displaying
#   a demo messages sequente (character sets etc.) for debugging the decoder.
#   For displaying live CC streams you can use the following:
#
#      ./capture.py --sliced | ./caption.py
#
#   The buttons on top of the GUI switch between Closed Caption channels 1-4
#   and Text channels 1-4.
#
#   (This script is a translation of test/caption.c in the libzvbi package,
#   albeit based on TkInter here.)

import sys
import argparse
import struct
import tkinter.font as tkf
from tkinter import *
import select
import Zvbi

# declare global variables
vbi = None
pgno = None
dx = None
infile = None
read_elapsed = 0
tk = None
canvas = None

# constants
DISP_WIDTH   = 640
DISP_HEIGHT  = 480
CELL_WIDTH   = 16
CELL_HEIGHT  = 26
DISP_X_OFF   = 40
DISP_Y_OFF   = 45

# hash array to hold IDs of rolling text rows in the canvas
shift = dict()
shift_step = 2

# canvas background color - visible where video would look through
COLORKEY = "#80FF80"

#
#  Remove one row's text content (i.e. all pixmaps in the row's area)
#
def draw_blank(row, col, n_cols):
    for cid in canvas.find( "overlapping",
                            col * CELL_WIDTH + 1 + DISP_X_OFF,
                            row * CELL_HEIGHT + 1 + DISP_Y_OFF,
                            (col + n_cols) * CELL_WIDTH - 2 + DISP_X_OFF,
                            (row + 1) * CELL_HEIGHT - 2 + DISP_X_OFF ):

        img = canvas.itemcget(cid, "image")

        # remove the pixmap from the canvas
        canvas.delete(cid)

        # destroy the image (important to free the image's memory)
        tk.call("image", "delete", img)


def is_transp(v):
    return ((v >> 16) & 0x0F) == Zvbi.VBI_TRANSPARENT_SPACE


#
#  Draw one row of text
#
def draw_row(pg, row):
    (rows, columns) = pg.get_page_size()
    prop = pg.get_page_text_properties()

    # first remove all old text in the row
    draw_blank(row, 0, columns)

    col = 0
    while col < columns:
        # skip transparent characters
        if is_transp(prop[row * columns + col]):
            col += 1
            continue

        # count number of subsequent non-transparent characters
        # (required as optimisation - drawing each char separately is very slow)
        i = col + 1
        while (i < columns) and not is_transp(prop[row * columns + i]):
            i += 1

        # create RGBA image of the character sequence
        fmt = Zvbi.VBI_PIXFMT_PAL8
        vbi_canvas = pg.draw_cc_page(col, row, i - col, 1, fmt=fmt)

        # convert into a pixmap via XPM
        img = tk.call("image", "create", "photo", "-data",
                      pg.canvas_to_ppm(vbi_canvas, fmt, aspect=1))

        # finally, display the pixmap in the canvas
        cid = canvas.create_image(col * CELL_WIDTH + DISP_X_OFF,
                                  row * CELL_HEIGHT + DISP_Y_OFF,
                                  anchor=NW, image=img)
        col = i


#
#  Timer event for rolling text rows
#
def bump(snap):
    renew = False

    for cid in list(shift.keys()):  # copy key list to allow modification during iteration
        d = shift[cid]
        if snap:
            step = d
        else:
            step = (d if (d < shift_step) else shift_step)

        canvas.move(cid, 0, 0 - step)

        shift[cid] -= step
        if shift[cid] <= 0:
            del shift[cid]
        else:
            renew = True

    if renew:
        tk.after(20 * shift_step, lambda: bump(0))


#
#  Scroll a range of rows upwards
#
def roll_up(pg, first_row, last_row):
    if True:  # ---- soft scrolling ----

        # snap possibly still moving rows into their target positions
        bump(1)

        for cid in canvas.find("overlapping",
                               0,
                               first_row * CELL_HEIGHT + 1 + DISP_Y_OFF,
                               DISP_WIDTH,
                               (last_row + 1) * CELL_HEIGHT - 1 + DISP_Y_OFF):
            shift[cid] = CELL_HEIGHT

            # start time to trigger smooth scrolling
            tk.after(20 + 20 * shift_step, lambda: bump(0))

    else: # ---- jumpy scrolling ----
        for cid in canvas.find("overlapping",
                               0,
                               first_row * CELL_HEIGHT + DISP_Y_OFF,
                               DISP_WIDTH,
                               (last_row + 1) * CELL_HEIGHT - 1 + DISP_Y_OFF):

            canvas.move(cid, 0, 0 - CELL_HEIGHT)


#
#  Update a range of text rows
#
def render(pg, y0, y1):
    # snap possibly still moving rows into their target positions
    bump(1)

    for row in range(y0, y1+1):
        draw_row(pg, row)


#
#  Clear all text on-screen
#
def clear():
    for cid in canvas.find("all"):
        img = canvas.itemcget(cid, "image")
        canvas.delete(cid)
        tk.call("image", "delete", img)


#
#  Callback invoked by the VBI decoder when a new CC line is available
#
def cc_handler(type, ev):
    if (pgno.get() != -1) and (ev.pgno != pgno.get()):
        return

    # Fetching & rendering in the handler
    # is a bad idea, but this is only a test

    pg = vbi.fetch_cc_page(ev.pgno)

    (rows, columns) = pg.get_page_size()
    (y0, y1, roll) = pg.get_page_dirty_range()

    if (abs (roll) > rows):
        clear()
    elif (roll == -1):
        #draw_blank(y0, 0, columns)
        roll_up(pg, y0+1, y1)
    else:
        render(pg, y0, y1)


#
#  Callback bound to CC channel changes
#
def reset():
    try:
        pg = vbi.fetch_cc_page(pgno.get())
        (rows, columns) = pg.get_page_size()
        render(pg, 0, rows - 1)
    except Zvbi.PageError:
        clear()


#
#  Create the GUI
#
def init_window():
    global tk
    global canvas
    global pgno

    tk = Tk()
    tk.wm_title('Caption decoder')

    pgno = IntVar(tk, 1)

    # at the top: button array to switch CC channels
    f = Frame(tk)
    b = Label(f, text="Page:")
    b.pack(side=LEFT)
    for i in range(1, 8+1):
        b = Radiobutton(f, text=i, value=i, variable=pgno, command=lambda: reset())
        b.pack(side=LEFT)

    f.pack(side=TOP)

    # canvas to display CC text (as pixmaps)
    canvas = Canvas(tk, borderwidth=1, relief="sunken",
                    background=COLORKEY,
                    height=DISP_HEIGHT, width=DISP_WIDTH)
    canvas.pack(side=TOP)
    canvas.focus_set()


#
#  Feed caption from live stream or file with sample data
#
def pes_mainloop():
    while True:
        buf = infile.read(2048)
        bytes_left = len(buf)
        if bytes_left == 0:
            break

        while (bytes_left > 0):
            n_lines = dx.cor(sliced, 64, pts, buf, bytes_left)
            if (n_lines > 0):
                vbi.decode_bytes(sliced, n_lines, pts / 90000.0)

        tk.after(20, pes_mainloop)
        return

    print("\rEnd of stream", file=sys.stderr)


#
#  Forward sliced data read from STDIN to service decoder
#
def old_mainloop():
    # avoid blocking in read when no data available: keep Tk mainloop alive
    if select.select([infile], [], [], 0)[0]:
        # one one frame's worth of sliced data from the input stream or file
        sl = read_sliced()
        if sl:
            (n_lines, timestamp, sliced) = sl

            # pack the read data into the normal slicer output format
            # (i.e. the format delivered by the librarie's internal slicer)
            REC_LEN = 4+4+56
            buf = bytearray(n_lines * REC_LEN)
            line_idx = 0
            for slc_id, line, data in sliced:
                struct.pack_into("=LL", buf, line_idx*REC_LEN, slc_id, line)
                d_idx = 0
                for d in data:
                    buf[line_idx*REC_LEN + 4+4 + d_idx] = d
                    d_idx += 1
                line_idx += 1

            # pass the full frame's data to the decoder
            vbi.decode_bytes(buf, n_lines, timestamp)

            tk.after(20, old_mainloop)

        else:
            print("\rEnd of stream", file=sys.stderr)
    else:
        tk.after(20, old_mainloop)


# ----------------------------------------------------------------------------
#
#  Generate artificial caption data
#
def cmd(n):
    global sim_buf
    global cmd_time

    sliced = struct.pack("=LLBB54x", Zvbi.VBI_SLICED_CAPTION_525,
                                     21,
                                     Zvbi.par8(n >> 8),
                                     Zvbi.par8(n & 0x7F))

    sim_buf.append(["sliced", sliced, cmd_time])
    #vbi.decode(sliced, 1, cmd_time)

    cmd_time += 1 / 29.97


def printc(v):
    global sim_buf

    cmd(v * 256 + 0x80)
    sim_buf.append(["delay", 1])


def prints(v):
    global sim_buf
    s = bytes(v, 'ascii')

    i = 0
    while i + 1 < len(s):
        cmd(s[i] * 256 + s[i+1])
        i += 2

    if i < len(s):
        cmd(s[i] * 256 + 0x80)

    sim_buf.append(["delay", 1])


# constants
white = 0
green = 1
red = 4
yellow = 5
blue = 2
cyan = 3
magenta = 6
black = 7

mapping_row = (2, 3, 4, 5,  10, 11, 12, 13, 14, 15,  0, 6, 7, 8, 9, -1)

italic = 7
underline = 1
opaque = 0
semi_transp = 1

# global variable
ch = 0

def BACKG(v,x):         cmd(0x2000); \
                        cmd(0x1020 + ((ch & 1) << 11) + (v << 1) + x)
def PREAMBLE(v,x,y): 
                        cmd(0x1040 + ((ch & 1) << 11) + ((mapping_row[v] & 14) << 7)
                             + ((mapping_row[v] & 1) << 5) + (x << 1) + y)
def INDENT(v,x,y): 
                        cmd(0x1050 + ((ch & 1) << 11) + ((mapping_row[v] & 14) << 7)
                                   + ((mapping_row[v] & 1) << 5) + ((x // 4) << 1) + y)
def MIDROW(v,x):        cmd(0x1120 + ((ch & 1) << 11) + (v << 1) + x)
def SPECIAL_CHAR(v):    cmd(0x1130 + ((ch & 1) << 11) + v)

def CCODE(v,x):         return (v + ((x & 1) << 11) + ((x & 2) << 7))
def RESUME_CAPTION():   cmd(CCODE(0x1420, ch))
def BACKSPACE():        cmd(CCODE(0x1421, ch))
def DELETE_EOR(v):      cmd(CCODE(0x1424, ch))
def ROLL_UP(v):         cmd(CCODE(0x1425, ch) + v - 2)
def FLASH_ON():         cmd(CCODE(0x1428, ch))
def RESUME_DIRECT():    cmd(CCODE(0x1429, ch))
def TEXT_RESTART():     cmd(CCODE(0x142A, ch))
def RESUME_TEXT():      cmd(CCODE(0x142B, ch))
def END_OF_CAPTION():   cmd(CCODE(0x142F, ch))
def ERASE_DISPLAY():    cmd(CCODE(0x142C, ch))
def CR():               cmd(CCODE(0x142D, ch))
def ERASE_HIDDEN():     cmd(CCODE(0x142E, ch))
def TAB(v):             cmd(CCODE(0x1720, ch) + v)
def TRANSP():           cmd(0x2000); cmd(0x172D + ((ch & 1) << 11))
def BLACK(v):           cmd(0x2000); cmd(0x172E + ((ch & 1) << 11) + v)

def PAUSE(n_frames):
    sim_buf.append(["delay", n_frames])


def hello_world():
    global ch
    global sim_buf
    global cmd_time

    sim_buf = []
    cmd_time = 0.0
    pgno.set(-1)

    prints(" HELLO WORLD! ")
    PAUSE(30)

    ch = 4
    TEXT_RESTART()
    prints("Character set - Text 1")
    CR(); CR()
    for i in range(32, 127+1):
        printc(i)
        if ((i & 15) == 15):
            CR()
    MIDROW(italic, 0)
    for i in range(32, 127+1):
        printc(i)
        if ((i & 15) == 15):
            CR()
    MIDROW(white, underline)
    for i in range(32, 127+1):
        printc(i)
        if ((i & 15) == 15):
            CR()
    MIDROW(white, 0)
    prints("Special: ")
    for i in range(0, 15+1):
        SPECIAL_CHAR(i)
    CR()
    prints("DONE - Text 1 ")
    PAUSE(50)

    ch = 5
    TEXT_RESTART()
    prints("Styles - Text 2")
    CR(); CR()
    MIDROW(white, 0); prints("WHITE"); CR()
    MIDROW(red, 0); prints("RED"); CR()
    MIDROW(green, 0); prints("GREEN"); CR()
    MIDROW(blue, 0); prints("BLUE"); CR()
    MIDROW(yellow, 0); prints("YELLOW"); CR()
    MIDROW(cyan, 0); prints("CYAN"); CR()
    MIDROW(magenta, 0); prints("MAGENTA"); BLACK (0); CR()
    BACKG(white, opaque); prints("WHITE"); BACKG (black, opaque); CR()
    BACKG(red, opaque); prints("RED"); BACKG (black, opaque); CR()
    BACKG(green, opaque); prints("GREEN"); BACKG (black, opaque); CR()
    BACKG(blue, opaque); prints("BLUE"); BACKG (black, opaque); CR()
    BACKG(yellow, opaque); prints("YELLOW"); BACKG (black, opaque); CR()
    BACKG(cyan, opaque); prints("CYAN"); BACKG (black, opaque); CR()
    BACKG(magenta, opaque); prints("MAGENTA"); BACKG (black, opaque); CR()
    PAUSE(200)
    TRANSP()
    prints(" TRANSPARENT BACKGROUND ")
    BACKG(black, opaque); CR()
    MIDROW(white, 0); FLASH_ON()
    prints(" Flashing Text  (if implemented) "); CR()
    MIDROW(white, 0); prints("DONE - Text 2 ")
    PAUSE(50)

    ch = 0
    ROLL_UP(2)
    ERASE_DISPLAY()
    prints(" ROLL-UP TEST "); CR(); PAUSE (20)
    prints("The ZVBI library provides"); CR(); PAUSE (20)
    prints("routines to access raw VBI"); CR(); PAUSE (20)
    prints("sampling devices (currently"); CR(); PAUSE (20)
    prints("the Linux V4L and and V4L2"); CR(); PAUSE (20)
    prints("API and the FreeBSD, OpenBSD,"); CR(); PAUSE (20)
    prints("NetBSD and BSDi bktr driver"); CR(); PAUSE (20)
    prints("API are supported), a versatile"); CR(); PAUSE (20)
    prints("raw VBI bit slicer, decoders"); CR(); PAUSE (20)
    prints("for various data services and"); CR(); PAUSE (20)
    prints("basic search, render and export"); CR(); PAUSE (20)
    prints("functions for text pages. The"); CR(); PAUSE (20)
    prints("library was written for the"); CR(); PAUSE (20)
    prints("Zapping TV viewer and Zapzilla"); CR(); PAUSE (20)
    prints("Teletext browser."); CR(); PAUSE (20)
    CR(); PAUSE(30)
    prints(" DONE - Caption 1 ")
    PAUSE(30)

    ch = 1
    RESUME_DIRECT()
    ERASE_DISPLAY()
    MIDROW(yellow, 0)
    INDENT(2, 10, 0); prints(" FOO "); CR()
    INDENT(3, 10, 0); prints(" MIKE WAS HERE "); CR(); PAUSE (20)
    MIDROW(red, 0)
    INDENT(6, 13, 0); prints(" AND NOW... "); CR()
    INDENT(8, 13, 0); prints(" TOM'S THERE TOO "); CR(); PAUSE (20)
    PREAMBLE(12, cyan, 0)
    prints("01234567890123456789012345678901234567890123456789"); CR()
    MIDROW(white, 0)
    prints(" DONE - Caption 2 "); CR()
    PAUSE(30)


#
#  Play back the buffered (simulated) CC data
#
def play_world():
    while len(sim_buf) > 0:
        a = sim_buf.pop(0)

        if (a[0] == "delay"):
            # delay event -> stop play-back and start timer
            tk.after(25 * a[1], play_world)
            break
        else:
            # pass VBI Data to the decoder context
            # (will trigger display via the CC callback function)
            vbi.decode_bytes(a[1], 1, a[2])


# ------- slicer.c -----------------------------------------------------------
#
# Read one frame's worth of sliced data (written by decode.pl)
# from a file or pipe (not used in demo mode)
#
def read_sliced():
    global read_elapsed

    # Read timestamp: newline-terminated string
    buf = ""
    while True:
        b = infile.read(1)
        if len(b) == 0:  # EOF
            return None
        c = "%c" % b[0]
        if c == '\n':
            break
        buf += c

    # Time in seconds since last frame.
    try:
        dt = float(buf)
        if (dt < 0.0):
            dt = -dt
    except:
        print("invalid timestamp in input: " % dt, file=sys.stderr)
        sys.exit(1)

    timestamp = read_elapsed
    read_elapsed += dt

    # Read number of sliced lines
    n_lines = infile.read(1)[0]
    assert(n_lines >= 0) # "invalid line count in input"

    # Read sliced data: each record consists of:
    # - type ID (0..7)
    # - line index (LSB)
    # - line index (MSB)
    # - data x N: where N is implied by type ID
    sliced = []
    for n in range(0, n_lines):
        index = infile.read(1)[0]
        assert(index >= 0) # "invalid index"

        line = (infile.read(1)[0]
                   + 256 * infile.read(1)[0]) & 0xFFF

        if index == 0:
            slc_id = Zvbi.VBI_SLICED_TELETEXT_B
            data = infile.read(42)

        elif index == 1:
            slc_id = Zvbi.VBI_SLICED_CAPTION_625; 
            data = infile.read(2)

        elif index == 2:
            slc_id = Zvbi.VBI_SLICED_VPS
            data = infile.read(13)

        elif index == 3:
            slc_id = Zvbi.VBI_SLICED_WSS_625; 
            data = infile.read(2)

        elif index == 4:
            slc_id = Zvbi.VBI_SLICED_WSS_CPR1204; 
            data = infile.read(3)

        elif index == 7:
            slc_id = Zvbi.VBI_SLICED_CAPTION_525; 
            data = infile.read(2)

        else:
            print("Oops! Unknown data type index in sliced VBI file", file=sys.stderr)
            exit(1)

        sliced.append([slc_id, line, data])

    return (n_lines, timestamp, sliced)


# ----------------------------------------------------------------------------

def main_func():
    global vbi
    global infile

    # create the GUI
    init_window()

    # create a decoder context and enable Closed Captioning decoding
    vbi = Zvbi.ServiceDec()
    vbi.event_handler_register(Zvbi.VBI_EVENT_CAPTION, cc_handler)

    if sys.stdin.isatty():
        # no file or stream on STDIN -> generate demo data
        print("No input provided via STDIN - playing back demo sequence", file=sys.stderr)
        hello_world()
        # start play back of the demo data (timer-based, to give control to the main loop below)
        play_world()

    else:
        infile = open(sys.stdin.fileno(), "rb")

        #c = ord(infile.read(1) or 1)  # TODO
        #infile.ungetc(c)
        #if (0 == c):
        #    dx = Zvbi.DvbDemux.pes_new()
        #    tk.after(20, pes_mainloop)
        #else:
        if True:
            # install timer to poll for incoming data
            tk.after(20, old_mainloop)

    # everything from here on is event driven
    tk.mainloop()

try:
    main_func()
except KeyboardInterrupt:
    pass



#!/usr/bin/python3
#
#  zvbi-decode -- decode sliced VBI data using low-level
#                  libzvbi functions
#
#  Copyright (C) 2004, 2006 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#/

# Description:
#
#   Example for the use of class Zvbi.ServiceDec. Decodes sliced VBI data
#   on stdin for different data services. Example for decoding teletext:
#
#     ./capture.py --sliced | ./decode.py --ttx
#
#   Call with option --help for a list of options.
#   (This is a direct translation of test/decode.c in the libzvbi package.)

import sys
import argparse
import struct
import re
import Zvbi

# Demultiplexers.
pfc = None
dvb = None
idl = None
xds = None

# Reader for old test/capture --sliced output.
infile = None
outfile = None
read_elapsed = False

# ------- slicer.c -----------------------------------------------------------
#
# Read one frame's worth of sliced data (written by decode.pl)
# from a file or pipe (not used in demo mode)
#
def read_sliced():
    global read_elapsed

    # Read timestamp: newline-terminated string
    buf = ""
    while True:
        b = infile.read(1)
        if len(b) == 0:  # EOF
            return None
        c = "%c" % b[0]
        if c == '\n':
            break
        buf += c

    # Time in seconds since last frame.
    try:
        dt = float(buf)
        if (dt < 0.0):
            dt = -dt
    except:
        print("invalid timestamp in input: " % dt, file=sys.stderr)
        sys.exit(1)

    timestamp = read_elapsed
    read_elapsed += dt

    # Read number of sliced lines
    n_lines = infile.read(1)[0]
    assert(n_lines >= 0) # "invalid line count in input"

    # Read sliced data: each record consists of:
    # - type ID (0..7)
    # - line index (LSB)
    # - line index (MSB)
    # - data x N: where N is implied by type ID
    sliced = []
    for n in range(0, n_lines):
        index = infile.read(1)[0]
        assert(index >= 0) # "invalid index"

        line = (infile.read(1)[0]
                   + 256 * infile.read(1)[0]) & 0xFFF

        if index == 0:
            slc_id = Zvbi.VBI_SLICED_TELETEXT_B
            data = infile.read(42)

        elif index == 1:
            slc_id = Zvbi.VBI_SLICED_CAPTION_625
            data = infile.read(2)

        elif index == 2:
            slc_id = Zvbi.VBI_SLICED_VPS
            data = infile.read(13)

        elif index == 3:
            slc_id = Zvbi.VBI_SLICED_WSS_625
            data = infile.read(2)

        elif index == 4:
            slc_id = Zvbi.VBI_SLICED_WSS_CPR1204
            data = infile.read(3)

        elif index == 7:
            slc_id = Zvbi.VBI_SLICED_CAPTION_525
            data = infile.read(2)

        else:
            print("Oops! Unknown data type index in sliced VBI file", file=sys.stderr)
            exit(1)

        sliced.append([data, slc_id, line])

    return (n_lines, timestamp, sliced)


def _vbi_pfc_block_dump(pgno, stream, app_id, block, binary):
    print("PFC pgno=%x stream=%u id=%u size=%u" %
          (pgno, stream, app_id, length(block)))

    if binary:
        # cannot mix binary write with text print within same stream in Python
        #sys.stdout.write(block)
        print(block)
    else:
        block = Zvbi.unpar_str(block, '.')
        block = re.sub(r'[\x00-\x1F\x7F]', '.', block.decode('ISO-8859-1'))
        # missing: insert \n every 75 chars

        print(block)


def put_cc_char(c1, c2):
    # All caption characters are representable in UTF-8
    c = ((c1 << 8) + c2) & 0x777F
    ucs2_str = Zvbi.caption_unicode(c)  # !to_upper

    print(ucs2_str)


def caption_command(line, c1, c2):
    print("CC line=%3u cmd 0x%02x 0x%02x " % (line, c1, c2), end='')

    if 0 == c1:
        print("null")
        return
    elif c2 < 0x20:
        print("invalid")
        return

    # Some common bit groups.

    ch = (c1 >> 3) & 1  # channel
    a7 = c1 & 7
    f = c1 & 1  # field
    b7 = (c2 >> 1) & 7
    u = c2 & 1  # underline

    if (c2 >= 0x40):
        row = (
               10,                # 0     # 0x1040
               -1,                # 1     # unassigned
               0, 1, 2, 3,        # 2     # 0x1140 ... 0x1260
               11, 12, 13, 14,    # 6     # 0x1340 ... 0x1460
               4, 5, 6, 7, 8, 9   # 10    # 0x1540 ... 0x1760
              )

        # Preamble Address Codes -- 001 crrr  1ri bbbu

        rrrr = a7 * 2 + ((c2 >> 5) & 1)

        if c2 & 0x1:
            print("PAC ch=%u row=%u column=%u u=%u" %
                    (ch, row[rrrr], b7 * 4, u))
        else:
            print("PAC ch=%u row=%u color=%u u=%u" %
                    (ch, row[rrrr], b7, u))
        return

    # Control codes -- 001 caaa  01x bbbu

    if a7 == 0:
        if (c2 < 0x30):
            mnemo_1 = (
                        "BWO", "BWS", "BGO", "BGS",
                        "BBO", "BBS", "BCO", "BCS",
                        "BRO", "BRS", "BYO", "BYS",
                        "BMO", "BMS", "BAO", "BAS"
                      )

            print("%s ch=%u" % (mnemo_1[c2 & 0xF], ch))
            return

    elif a7 == 1:
            if (c2 < 0x30):
                print("mid-row ch=%u color=%u u=%u" % (ch, b7, u), end='')
            else:
                print("special character ch=%u 0x%02x%02x='" % (ch, c1, c2), end='')
                put_cc_char(c1, c2)
                print("'")
            return

    elif a7 in (2,3):  # first & second group
        print("extended character ch=%u 0x%02x%02x='" % (ch, c1, c2), end='')
        put_cc_char(c1, c2)
        print("'")
        return

    elif a7 in (4,5): # f=0,1
        if (c2 < 0x30):
            mnemo_2 = (
                        "RCL", "BS",  "AOF", "AON",
                        "DER", "RU2", "RU3", "RU4",
                        "FON", "RDC", "TR",  "RTD",
                        "EDM", "CR",  "ENM", "EOC"
                      )

            print("%s ch=%u f=%u" % (mnemo_2[c2 & 0xF], ch, f))
            return

    elif a7 == 6: # reserved
        pass

    elif a7 == 7:
        if c2 == 0x21 or c2 == 0x22 or c2 == 0x23:
            print("TO%u ch=%u" % (c2 - 0x20, ch))
            return

        elif c2 == 0x2D:
            print("BT ch=%u" % ch)
            return

        elif c2 == 0x2E:
            print("FA ch=%u" % ch)
            return

        elif c2 == 0x2F:
            print("FAU ch=%u" % ch)
            return

    print("unknown")


def xds_cb(xds_class, xds_subclass, buf, user_data=None):
    #_vbi_xds_packet_dump(xp, stdout)
    print("XDS packet:", xds_class, xds_subclass, buf)

    return True  # no errors


def caption(inbuf, line):
    buffer = inbuf[0 : 1]

    if (opt.decode_xds and 284 == line):
        try:
            xds.feed(inbuf)
        except Zvbi.XdsDemuxError:
            print("Parity error in XDS data.")

    if (opt.decode_caption
        and (21 == line or 284 == line # NTSC
            or 22 == line)):

        c1 = Zvbi.unpar8(buffer[0])
        c2 = Zvbi.unpar8(buffer[1])

        if ((c1 | c2) < 0):
            print("Parity error in CC line=%u %s0x%02x %s0x%02x." %
                    (line,
                     (">" if (c1 < 0) else ""), buffer[0] & 0xFF,
                     (">" if (c2 < 0) else ""), buffer[1] & 0xFF))

        elif (c1 >= 0x20):
            print("CC line=%3u text 0x%02x 0x%02x '" %
                    (line, c1, c2), end='')

            # All caption characters are representable
            # in UTF-8, but not necessarily in ASCII.
            text = struct.pack("2B", c1, c2)

            # Error ignored.
            utf = Zvbi.iconv_caption(text, ord('?'))
            # suppress warnings about wide characters
            #utf = encode("ISO-8859-1", utf, Encode::FB_DEFAULT)

            print(utf + "'")

        elif (0 == c1 or c1 >= 0x10):
            caption_command(line, c1, c2)

        elif opt.decode_xds:
            print("CC line=%3u cmd 0x%02x 0x%02x " % (line, c1, c2), end='')
            if (0x0F == c1):
                print("XDS packet end", end='')
            else:
                print("XDS packet start/continue", end='')


#if 3 == VBI_VERSION_MINOR # XXX port me back
#
#static void
#dump_cni                        (vbi_cni_type           type,
#                                 unsigned int           cni)
#{
#        vbi_network nk
#        vbi_bool success
#
#        if (!opt.dump_network)
#                return
#
#        success = vbi_network_init (&nk)
#        if (!success)
#                no_mem_exit ()
#
#        success = vbi_network_set_cni (&nk, type, cni)
#        if (!success)
#                no_mem_exit ()
#
#        _vbi_network_dump (&nk, stdout)
#        putchar ('\n')
#
#        vbi_network_destroy (&nk)
#}
#
#endif # 3 == VBI_VERSION_MINOR

def dump_bytes(buffer, n_bytes):
    if opt.dump_bin:
        outfile.write(buffer, n_bytes)
        return

    if opt.dump_hex:
        hl = (("%02x" % x) for x in buffer)
        print(" ".join(hl))

    # For Teletext: Not all characters are representable
    # in ASCII or even UTF-8, but at this stage we don't
    # know the Teletext code page for a proper conversion.
    astr = Zvbi.unpar_str(buffer, '.')
    astr = re.sub(r'[\x00-\x1F\x7F]', '.', astr.decode('ISO-8859-1'))

    print(">"+ astr[0 : n_bytes] +"<")

#if 3 == VBI_VERSION_MINOR # XXX port me back
#
#static void
#packet_8301                     (const uint8_t          buffer[42],
#                                 unsigned int           designation)
#{
#        unsigned int cni
#        time_t time
#        int gmtoff
#        struct tm tm
#
#        if (!opt.decode_8301)
#                return
#
#        if (!vbi_decode_teletext_8301_cni (&cni, buffer)) {
#                printf (_("Error in Teletext "
#                          "packet 8/30 format 1 CNI.\n"))
#                return
#        }
#
#        if (!vbi_decode_teletext_8301_local_time (&time, &gmtoff, buffer)) {
#                printf (_("Error in Teletext "
#                          "packet 8/30 format 1 local time.\n"))
#                return
#        }
#
#        printf ("Teletext packet 8/30/%u cni=%x time=%u gmtoff=%d ",
#                designation, cni, (unsigned int) time, gmtoff)
#
#        gmtime_r (&time, &tm)
#
#        printf ("(%4u-%02u-%02u %02u:%02u:%02u UTC)\n",
#                tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
#                tm.tm_hour, tm.tm_min, tm.tm_sec)
#
#        if (0 != cni)
#                dump_cni (VBI_CNI_TYPE_8301, cni)
#}
#
#static void
#packet_8302                     (const uint8_t          buffer[42],
#                                 unsigned int           designation)
#{
#        unsigned int cni
#        vbi_program_id pi
#
#        if (!opt.decode_8302)
#                return
#
#        if (!vbi_decode_teletext_8302_cni (&cni, buffer)) {
#                printf (_("Error in Teletext "
#                          "packet 8/30 format 2 CNI.\n"))
#                return
#        }
#
#        if (!vbi_decode_teletext_8302_pdc (&pi, buffer)) {
#                printf (_("Error in Teletext "
#                          "packet 8/30 format 2 PDC data.\n"))
#                return
#        }
#
#        printf ("Teletext packet 8/30/%u cni=%x ", designation, cni)
#
#        _vbi_program_id_dump (&pi, stdout)
#
#        putchar ('\n')
#
#        if (0 != pi.cni)
#                dump_cni (pi.cni_type, pi.cni)
#}
#
#endif # 3 == VBI_VERSION_MINOR

def page_function_clear_cb(pgno, stream, app_id, block, user_data=None):
    _vbi_pfc_block_dump(pgno, stream, app_id, block, opt.dump_bin)
    return True


def  idl_format_a_cb(buffer, flags, user_data=None):
    if not opt.dump_bin:
        print("IDL-A%s%s " %
                (" <data lost>" if (flags & Zvbi.VBI_IDL_DATA_LOST) else ""),
                (" <dependent>" if (flags & Zvbi.VBI_IDL_DEPENDENT) else ""), end='')

    dump_bytes(buffer, len(buffer))

    return True


def calc_spa(spa_length, ord_buf):
    spa = 0

    for i in range(0, spa_length):
        h = Zvbi.unham8(ord_buf[4 + i])
        spa |= (h << (4 * i))

    return spa


def packet_idl(buffer, channel):
    ord_buf = buffer[0 : 10]

    print("IDL ch=%u " % channel, end='')

    if channel == 0:
        print("IDL: unexpected channel 0")
        exit(1)

    elif (channel >= 4 and channel <= 12):
        print("(Low bit rate audio) ", end='')

        dump_bytes(buffer, 42)

    elif channel in (5, 6, 13, 14):
        pa = Zvbi.unham8(ord_buf[3])
        pa |= Zvbi.unham8(ord_buf[4]) << 4
        pa |= Zvbi.unham8(ord_buf[5]) << 8

        if (pa < 0):
            print("Hamming error in Datavideo packet-address byte.")
            return

        print("(Datavideo) pa=0x%x " % pa, end='')

        dump_bytes(buffer, 42)

    elif channel in (8, 9, 10, 11, 15):
        # ft: format type
        ft = Zvbi.unham8(ord_buf[2])
        if (ft < 0):
            print("Hamming error in IDL format " +
                  "A or B format-type byte.")
            return

        if (0 == (ft & 1)):
            # ial: interpretation and address length
            # spa: service packet address

            ial = Zvbi.unham8(ord_buf[3])
            if (ial < 0):
                print("Hamming error in IDL format " +
                          "A interpretation-and-address-" +
                          "length byte.")
                return

            spa_length = ial & 7
            if (7 == spa_length):
                print("(Format A?) ", end='')
                dump_bytes(buffer, 42)
                return

            spa = calc_spa(spa_length, ord_buf)

            if (spa < 0):
                print("Hamming error in IDL format" +
                        "A service-packet-address byte.")
                return

            print("(Format A) spa=0x%x " % spa, end='')

        elif (1 == (ft & 3)):
            # an: application number
            # ai: application identifier

            an = (ft >> 2)

            ai = Zvbi.unham8(ord_buf[3])
            if (ai < 0):
                print("Hamming error in IDL format " +
                      "B application-number byte.")
                return

            print("(Format B) an=%d ai=%d " % (an, ai), end='')

        dump_bytes(buffer, 42)

    else:
        dump_bytes(buffer, 42)


def teletext(buffer, line):
    ord_buf = buffer[0 : 42]

    if not pfc is None:
        try:
            pfc.feed(buffer)
        except Zvbi.PfcDemuxError:
            print("Error in Teletext PFC packet.")
            return

    if not (opt.decode_ttx or
            opt.decode_8301 or
            opt.decode_8302 or
            opt.decode_idl):
        return

    pmag = Zvbi.unham16p(buffer)
    if pmag < 0:
        print("Hamming error in Teletext packet number.")
        return

    magazine = pmag & 7
    if (0 == magazine):
        magazine = 8
    packet = pmag >> 3

    if (8 == magazine and 30 == packet):
        designation = Zvbi.unham8(ord_buf[2])
        if (designation < 0 ):
            print("Hamming error in Teletext packet 8/30 designation byte.")
            return

        if (designation >= 0 and designation <= 1):
#if 3 == VBI_VERSION_MINOR # XXX port me back
            #packet_8301(buffer, designation)
#endif
            return

        if (designation >= 2 and designation <= 3):
#if 3 == VBI_VERSION_MINOR # XXX port me back
            #packet_8302(buffer, designation)
#endif
            return

    if (30 == packet or 31 == packet):
        if opt.decode_idl:
            packet_idl(buffer, pmag & 15)
            #print("Teletext IDL packet %u/%2u " %(magazine, packet), end='')
            #dump_bytes(buffer, 42)
            return

    if opt.decode_ttx:
            print("Teletext line=%3u %x/%2u " %
                    (line, magazine, packet), end='')
            dump_bytes(buffer, 42)
            return

vps_pr_label = ["", ""]
vps_label = [bytearray([ord(' ')] * 16), bytearray([ord(' ')] * 16)]
vps_label_off = [0, 0]

def vps(inbuf, line):
    global vps_pr_label
    global vps_label
    global vps_label_off

    if opt.decode_vps:
        if opt.dump_bin:
            print("VPS line=%3u " % line, end='')
            outfile.write(inbuf, 13)
            return

        try:
            cni = Zvbi.decode_vps_cni(inbuf)
        except ZvbiError:
            print("Error in VPS packet CNI.")
            return

#if 3 == VBI_VERSION_MINOR
#           if (!vbi_decode_vps_pdc(&pi, buffer)) {
#                printf "Error in VPS packet PDC data.\n"
#                return
#
#           printf "VPS line=%3u ", line
#
#           _vbi_program_id_dump(&pi, stdout)
#
#           putchar ('\n')
#
#           if (0 != pi.cni)
#               dump_cni(pi.cni_type, pi.cni)
#else
        print("VPS line=%3u CNI=%x" % (line, cni))
#endif

    if opt.decode_vps_other:
        l = (line != 16)

        c = Zvbi.rev8(inbuf[1])

        if (c & 0x80):
            vps_pr_label[l] = vps_label[l][0 : vps_label_off[l]].decode('ascii')
            vps_label_off[l] = 0

        cp = c & 0x7F
        if (cp < 0x20 or cp >= 0x7f):
            cp = ord('.')

        vps_label[l][vps_label_off[l]] = cp

        vps_label_off[l] = (vps_label_off[l] + 1) % 16

        print(("VPS line=%3u bytes 3-10: " +
                "%02x %02x (%02x='%c') %02x %02x " +
                "%02x %02x %02x %02x (\"%s\")") %
               (line,
                inbuf[0], inbuf[1],
                c, cp,
                inbuf[2], inbuf[3],
                inbuf[4], inbuf[5], inbuf[6], inbuf[7],
                vps_pr_label[l]))


#if 3 == VBI_VERSION_MINOR # XXX port me back
#
#static void
#wss_625                         (const uint8_t          buffer[2])
#{
#        if (opt.decode_wss) {
#                vbi_aspect_ratio ar
#
#                if (!vbi_decode_wss_625(&ar, buffer)) {
#                        printf (_("Error in WSS packet.\n"))
#                        return
#                }
#
#                fputs ("WSS ", stdout)
#
#                _vbi_aspect_ratio_dump(&ar, stdout)
#
#                putchar ('\n')
#        }
#}
#
#endif # 3 == VBI_VERSION_MINOR

last_sample_time = 0.0
last_stream_time = 0

def decode_vbi(sliced, n_lines, sample_time, stream_time):
    global last_sample_time
    global last_stream_time

    if opt.dump_time:
        # Sample time: When we captured the data, in
        #              seconds since 1970-01-01 (gettimeofday()).
        # Stream time: For ATSC/DVB the Presentation TimeStamp.
        #              For analog the frame number multiplied by
        #              the nominal frame period (1/25 or
        #              1001/30000 s). Both given in 90000 kHz units.
        # Note this isn't fully implemented yet. */
        print("ST %f (%+f) PTS %lld (%+lld)" %
                (sample_time, sample_time - last_sample_time,
                 stream_time, stream_time - last_stream_time))

        last_sample_time = sample_time
        last_stream_time = stream_time

    for i in range(n_lines):
        (data, slc_id, line) = sliced[i]

        if ( (slc_id == Zvbi.VBI_SLICED_TELETEXT_B_L10_625) or
             (slc_id == Zvbi.VBI_SLICED_TELETEXT_B_L25_625) or
             (slc_id == Zvbi.VBI_SLICED_TELETEXT_B_625) ):
            teletext(data, line)

        elif ( (slc_id == Zvbi.VBI_SLICED_VPS) or
               (slc_id == Zvbi.VBI_SLICED_VPS_F2) ):
            vps(data, line)

        elif ( (slc_id == Zvbi.VBI_SLICED_CAPTION_625_F1) or
               (slc_id == Zvbi.VBI_SLICED_CAPTION_625_F2) or
               (slc_id == Zvbi.VBI_SLICED_CAPTION_625) or
               (slc_id == Zvbi.VBI_SLICED_CAPTION_525_F1) or
               (slc_id == Zvbi.VBI_SLICED_CAPTION_525_F2) or
               (slc_id == Zvbi.VBI_SLICED_CAPTION_525) ):
            caption(data, line)

        elif (slc_id == Zvbi.VBI_SLICED_WSS_625):
            #3 wss_625(data)
            pass

        elif (slc_id == Zvbi.VBI_SLICED_WSS_CPR1204):
            pass


def dvb_feed_cb(sliced_buf, n_lines, pts):
    if n_lines > 0:
        # pull all data lines out of the packed slicer buffer
        # since we want to process them by Python code
        # (something we'd normally like to avoid, as it's slow)
        # (see export.pl for an efficient use case)
        sliced = list(sliced_buf)

        decode_vbi(sliced, n_lines,
                   0,    # sample_time
                   pts)  # stream_time

    # return TRUE in case we're invoked as callback via feed()
    return True


def pes_mainloop():
    while True:
        buf = infile.read(2048)
        if len(buf) == 0:
            break

        dvb.feed(buf)
        for sliced_buf in dvb:
            dvb_feed_cb(sliced_buf, len(sliced_buf), sliced_buf.timestamp)

    print("\rEnd of stream", file=sys.stderr)


def old_mainloop():
    # one one frame's worth of sliced data from the input stream or file
    while True:
        sl = read_sliced()
        if sl:
            (n_lines, timestamp, sliced) = sl

            # pass the full frame's data to the decoder
            decode_vbi(sliced, n_lines, timestamp, 0)
        else:
            break

    print("\rEnd of stream", file=sys.stderr)


def ttx_pgno(val_str):
    val = int(val_str, 16)
    if (val < 0x100) or (val > 0x8FF):
        raise ValueError("Teletext page not in range 100-8FF")
    return val


def main_func():
    global pfc
    global idl
    global xds
    global dvb
    global infile
    global outfile

    if opt.all:
        opt.decode_ttx = True
        opt.decode_8301 = True
        opt.decode_8302 = True
        opt.decode_caption = True
        opt.decode_vps = True
        opt.decode_wss = True
        opt.decode_idl = True
        opt.decode_xds = True
        opt.pfc_pgno = 0x1DF

    if sys.stdin.isatty():
        print("No VBI data on standard input.", file=sys.stderr)
        sys.exit(1)

    if opt.pfc_pgno:
        pfc = Zvbi.PfcDemux(opt.pfc_pgno,
                            opt.pfc_stream,
                            page_function_clear_cb)

    if opt.idl_channel:
        idl = Zvbi.IdlDemux(opt.idl_channel,
                            opt.idl_address,
                            idl_format_a_cb)

    if opt.decode_xds:
        xds = Zvbi.XdsDemux(xds_cb)

    if opt.verbose:
        Zvbi.set_log_on_stderr(-1)

    outfile = open(sys.stdout.fileno(), "wb")
    infile = open(sys.stdin.fileno(), "rb")

    #c = ord(infile.read(1) || 1)
    #infile.fseek(0, 0)
    c=1

    if (0 == c or opt.source_is_pes):
        #dvb = Zvbi.DvbDemux(dvb_feed_cb)
        dvb = Zvbi.DvbDemux()
        if opt.verbose:
            dvb.set_log_fn
            dvb.set_log_fn(Zvbi.VBI_LOG_DEBUG, lambda l,c,m: print("DEMUX LOG", l, c, m, file=sys.stderr))
        pes_mainloop()
    else:
#if 2 == VBI_VERSION_MINOR # XXX port me
        #open_sliced_read(infile)
#endif
        old_mainloop()


def ParseCmdOptions():
    global opt
    desc = "ZVBI decoding examples\n" + \
           "Copyright (C) 2004, 2006 Michael H. Schimek\n" + \
           "This program is licensed under GPL 2 or later. NO WARRANTIES."
    parser = argparse.ArgumentParser(description=desc)
    # Input options:
    parser.add_argument("--pes", action='store_true', default=False, dest="source_is_pes", help="Source is a DVB PES stream [auto-detected]") # ATSC/DVB
    # Decoding options:
    parser.add_argument("--8301", action='store_true', default=False, dest="decode_8301", help="Teletext packet 8/30 format 1 (local time)")
    parser.add_argument("--8302", action='store_true', default=False, dest="decode_8302", help="Teletext packet 8/30 format 2 (PDC)")
    parser.add_argument("--all", action='store_true', default=False, dest="all", help="Enable ttx,8301,8302,cc,idl,vps,wssxds,pfc")
    parser.add_argument("--cc", action='store_true', default=False, dest="decode_caption", help="Closed Caption")
    parser.add_argument("--idl", action='store_true', default=False, dest="decode_idl", help="Any Teletext IDL packets (M/30, M/31)")
    parser.add_argument("--idl-ch", type=int, default=0, dest="idl_channel", help="Decode Teletext IDL format A data from channel N")
    parser.add_argument("--idl-addr", type=int, default=0, dest="idl_address", help="Decode Teletext IDL format A data from address NNN")
    parser.add_argument("--pfc-pgno", type=ttx_pgno, default=0, dest="pfc_pgno", help="Decode Teletext Page Function Clear data from page NNN (e.g. 1DF)")
    parser.add_argument("--pfc-stream", type=int, default=0, dest="pfc_stream", help="Decode Teletext Page Function Clear data from stream NN")
    parser.add_argument("--vps", action='store_true', default=False, dest="decode_vps", help="Decode VPS data unrelated to PDC")
    parser.add_argument("--vps-other", action='store_true', default=False, dest="decode_vps_other", help="Decode VPS data unrelated to PDC")
    parser.add_argument("--ttx", action='store_true', default=False, dest="decode_ttx", help="Decode any Teletext packet")
    parser.add_argument("--wss", action='store_true', default=False, dest="decode_wss", help="Wide Screen Signalling")
    parser.add_argument("--xds", action='store_true', default=False, dest="decode_xds", help="Decode eXtended Data Service (NTSC line 284)")
    # Modifying options:
    parser.add_argument("--hex", action='store_true', default=False, dest="dump_hex", help="With -t dump packets in hex and ASCII, otherwise only ASCII")
    parser.add_argument("--network", action='store_true', default=False, dest="dump_network", help="With -1, -2, -v decode CNI and print available information about the network")
    parser.add_argument("--bin", action='store_true', default=False, dest="dump_bin", help="With -t, -p, -v dump data in binary format instead of ASCII")
    parser.add_argument("--time", action='store_true', default=False, dest="dump_time", help="Dump capture timestamps")
    # misc
    parser.add_argument("--verbose", action='store_true', default=False, help="Enable trace in VBI library")
    parser.add_argument("--version", action='store_true', default=False, dest="version", help="Print the program version and exit")
    opt = parser.parse_args()



try:
    ParseCmdOptions()
    main_func()
except KeyboardInterrupt:
    pass



#!/usr/bin/python3
#
#  libzvbi test
#  Copyright (C) 2006 Michael H. Schimek
#  Perl Port: Copyright (C) 2007 Tom Zoerner
#  Python Port: Copyright (C) 2020 Tom Zoerner
#
# Description:
#
#   This script contains tests for encoding and decoding the VPS data
#   service on randomly generated data.
#
#   (This is a direct translation of test/test-vps.c in libzvbi.)

import random
import Zvbi

def test_invalid(cni):
    try:
        foo = Zvbi.encode_vps_cni(cni)
        raise Exception("Unexpected success encoding VPS CNI 0x%X" % cni)
    except Zvbi.Error:
        pass

def main():
    cnis = [0x000, 0x001, 0x5A5, 0xA5A, 0xFFF]

    random.seed()
    rands = bytearray()
    for i in range(0, 13):
        rands.append(random.randrange(0, 256))

    buffer_dec = bytes(rands)
    buffer_enc = bytes(rands)

    cni2 = Zvbi.decode_vps_cni(buffer_dec)

    buffer_enc = Zvbi.encode_vps_cni(cni2)
    #if not buffer_enc == buffer_dec:
    #    print("ERROR: mismatch")

    for enc_cni in cnis:
        buffer_enc = Zvbi.encode_vps_cni(enc_cni)

        dec_cni = Zvbi.decode_vps_cni(buffer_enc)
        if not dec_cni == enc_cni:
            raise Exception("ERROR: mismatch CNI 0x%X -> 0x%X" % (enc_cni, dec_cni))

    test_invalid(-1)
    test_invalid(0x1000)
    test_invalid((1<<31)-1)

    print("OK")

main()
